/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

// Recherche d'un motif dans un texte avec l'algorithme de Boyer--Moore

#include "prelude.h"

// Avec 256 caractères, on adopte ici une représentation simple pour
// la table des décalages, avec un tableau de tableaux (avec plus de
// caractères, il faudrait prendre un dictionnaire pour chaque ligne,
// une table de hachage ou un arbre binaire par exemple).

// Les entrées vides de la table sont préremplies avec -1, ce qui
// simplifie le code de la fonction de recherche `boyer_moore`
// ci-dessous.

// Note : dans les deux fonctions ci-dessous, on prend soin de
// convertir les caractères vers le type `unsigned char` (car le
// caractère signé ou non du type `char` est défini seulement par
// l'implémentation).

// construit la table pour le motif `m` de longueur `lm`
int **build_table(char *m, int lm) {
  int **table = calloc(lm, sizeof(int*));
  for (int j = 0; j < lm; j++) {
    table[j] = calloc(256, sizeof(int));
    for (int c = 0; c < 256; c++)
      table[j][c] = -1;
    for (int k = 0; k < j; k++) {
      unsigned char c = m[k];
      table[j][c] = k;
    }
  }
  return table;
}

// affiche toutes les occurrences de `m` dans `t` avec Boyer--Moore
// et renvoie le nombre d'occurrences
int boyer_moore(char *m, char *t) {
  int lm = strlen(m), lt = strlen(t);
  int **table = build_table(m, lm);
  int count = 0;
  for (int i = 0; i <= lt - lm;) {
    int k = 0;
    for (int j = lm - 1; j >= 0; j--) {
      if (t[i + j] != m[j]) {
        unsigned char c = t[i + j];
        k = j - table[j][c];
        break;
      }
    }
    if (k == 0) {
      printf("occurrence à la position %d\n", i);
      count++;
      k = 1;
    }
    i += k;
  }
  return count;
}

// variante utilisant plutôt la bibliothèque matrix
#include "matrix.h"

matrix *build_table_matrix(char *m) {
  int lm = strlen(m);
  matrix *table = matrix_create(lm, 256, -1);
  for (int j = 0; j < lm; j++)
    for (int k = 0; k < j; k++)
      matrix_set(table, j, m[k], k);
  return table;
}

int boyer_moore_matrix(char *m, char *t) {
  int lm = strlen(m), lt = strlen(t);
  matrix *table = build_table_matrix(m);
  int count = 0;
  for (int i = 0; i <= lt - lm;) {
    int k = 0;
    for (int j = lm - 1; j >= 0; j--) {
      if (t[i + j] != m[j]) {
        k = j - matrix_get(table, j, t[i + j]);
        break;
      }
    }
    if (k == 0) {
      printf("occurrence à la position %d\n", i);
      count++;
      k = 1;
    }
    i += k;
  }
  return count;
}
